@@ -158,8 +158,9 @@ struct ykpiv_state {
   SCARDHANDLE card;
   int  verbose;
   char *pin;
+  uint8_t *mgm_key;
   ykpiv_allocator allocator;
-  bool isNEO;
+  uint32_t model;
   ykpiv_version_t ver;
   uint32_t serial;
 };
