@@ -57,12 +57,13 @@
  */
 const uint8_t CHUID_TMPL[] = {
   0x30, 0x19, 0xd4, 0xe7, 0x39, 0xda, 0x73, 0x9c, 0xed, 0x39, 0xce, 0x73, 0x9d,
-  0x83, 0x68, 0x58, 0x21, 0x08, 0x42, 0x10, 0x84, 0x21, 0x38, 0x42, 0x10, 0xc3,
-  0xf5, 0x34, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+  0x83, 0x68, 0x58, 0x21, 0x08, 0x42, 0x10, 0x84, 0x21, 0xc8, 0x42, 0x10, 0xc3,
+  0xeb, 0x34, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x35, 0x08, 0x32, 0x30, 0x33, 0x30, 0x30,
   0x31, 0x30, 0x31, 0x3e, 0x00, 0xfe, 0x00,
 };
 #define CHUID_GUID_OFFS 29
+#define TAG_CHUID_UUID 0x34
 
 // f0: Card Identifier
 //  - 0xa000000116 == GSC-IS RID
@@ -86,7 +87,7 @@ static ykpiv_rc _get_metadata_item(uint8_t *data, size_t cb_data, uint8_t tag, u
 static ykpiv_rc _set_metadata_item(uint8_t *data, size_t *pcb_data, size_t cb_data_max, uint8_t tag, uint8_t *p_item, size_t cb_item);
 
 static size_t _obj_size_max(ykpiv_state *state) {
-  return (state && state->isNEO) ? CB_OBJ_MAX_NEO : CB_OBJ_MAX;
+  return (state && state->model == DEVTYPE_NEOr3) ? CB_OBJ_MAX_NEO : CB_OBJ_MAX;
 }
 
 /*
@@ -96,21 +97,46 @@ static size_t _obj_size_max(ykpiv_state *state) {
 ykpiv_rc ykpiv_util_get_cardid(ykpiv_state *state, ykpiv_cardid *cardid) {
   ykpiv_rc res = YKPIV_OK;
   uint8_t buf[CB_OBJ_MAX];
-  size_t len = sizeof(buf);
+  unsigned long len = sizeof(buf);
+  uint8_t *p_temp = NULL;
+  size_t cb_temp = 0;
+  uint8_t tag = 0;
 
   if (!cardid) return YKPIV_GENERIC_ERROR;
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
-  res = _ykpiv_fetch_object(state, YKPIV_OBJ_CHUID, buf, (unsigned long *)&len);
-  if (YKPIV_OK == res) {
-    if (len != sizeof(CHUID_TMPL)) {
-      res = YKPIV_GENERIC_ERROR;
-    }
-    else {
-      memcpy(cardid->data, buf + CHUID_GUID_OFFS, YKPIV_CARDID_SIZE);
+  if ((res = _ykpiv_fetch_object(state, YKPIV_OBJ_CHUID, buf, &len)) == YKPIV_OK) {
+    p_temp = buf;
+
+    while (p_temp < (buf + len)) {
+      tag = *p_temp++;
+
+      if (!_ykpiv_has_valid_length(p_temp, (size_t)(buf + len - p_temp))) {
+        res = YKPIV_SIZE_ERROR;
+        goto Cleanup;
+      }
+
+      p_temp += _ykpiv_get_length(p_temp, &cb_temp);
+
+      if (tag == TAG_CHUID_UUID) {
+        /* found card uuid */
+        if (cb_temp < YKPIV_CARDID_SIZE || p_temp + YKPIV_CARDID_SIZE > buf + len) {
+          res = YKPIV_SIZE_ERROR;
+          goto Cleanup;
+        }
+
+        res = YKPIV_OK;
+        memcpy(cardid->data, p_temp, YKPIV_CARDID_SIZE);
+        goto Cleanup;
+      }
+
+      p_temp += cb_temp;
     }
+
+    /* not found, not malformed */
+    res = YKPIV_GENERIC_ERROR;
   }
 
 Cleanup:
@@ -136,7 +162,7 @@ ykpiv_rc ykpiv_util_set_cardid(ykpiv_state *state, const ykpiv_cardid *cardid) {
     memcpy(id, cardid->data, sizeof(id));
   }
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   memcpy(buf, CHUID_TMPL, sizeof(CHUID_TMPL));
@@ -154,14 +180,14 @@ Cleanup:
 ykpiv_rc ykpiv_util_get_cccid(ykpiv_state *state, ykpiv_cccid *ccc) {
   ykpiv_rc res = YKPIV_OK;
   uint8_t buf[CB_OBJ_MAX];
-  size_t len = sizeof(buf);
+  unsigned long len = sizeof(buf);
 
   if (!ccc) return YKPIV_GENERIC_ERROR;
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
-  res = _ykpiv_fetch_object(state, YKPIV_OBJ_CAPABILITY, buf, (unsigned long *)&len);
+  res = _ykpiv_fetch_object(state, YKPIV_OBJ_CAPABILITY, buf, &len);
   if (YKPIV_OK == res) {
     if (len != sizeof(CCC_TMPL)) {
       res = YKPIV_GENERIC_ERROR;
@@ -194,7 +220,7 @@ ykpiv_rc ykpiv_util_set_cccid(ykpiv_state *state, const ykpiv_cccid *ccc) {
     memcpy(id, ccc->data, sizeof(id));
   }
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   len = sizeof(CCC_TMPL);
@@ -208,20 +234,21 @@ Cleanup:
 }
 
 ykpiv_devmodel ykpiv_util_devicemodel(ykpiv_state *state) {
-  if (!state || state->context == SCARD_E_INVALID_HANDLE)
+  if (!state || !state->context || (state->context == (SCARDCONTEXT)-1)) {
     return DEVTYPE_UNKNOWN;
-  return (state->isNEO ? DEVTYPE_NEOr3 : DEVTYPE_YK4);
+  }
+  return state->model;
 }
 
 ykpiv_rc ykpiv_util_list_keys(ykpiv_state *state, uint8_t *key_count, ykpiv_key **data, size_t *data_len) {
   ykpiv_rc res = YKPIV_OK;
   ykpiv_key *pKey = NULL;
   uint8_t *pData = NULL;
+  uint8_t *pTemp = NULL;
   size_t cbData = 0;
   size_t offset = 0;
   uint8_t buf[CB_BUF_MAX];
   size_t cbBuf = 0;
-  bool transaction = false;
   size_t i = 0;
   size_t cbRealloc = 0;
 
@@ -256,7 +283,7 @@ ykpiv_rc ykpiv_util_list_keys(ykpiv_state *state, uint8_t *key_count, ykpiv_key
 
   if ((NULL == data) || (NULL == data_len) || (NULL == key_count)) { return YKPIV_GENERIC_ERROR; }
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   // init return parameters
@@ -274,17 +301,21 @@ ykpiv_rc ykpiv_util_list_keys(ykpiv_state *state, uint8_t *key_count, ykpiv_key
 
   for (i = 0; i < sizeof(SLOTS); i++) {
     cbBuf = sizeof(buf);
+    res = _read_certificate(state, SLOTS[i], buf, &cbBuf);
 
-    if (YKPIV_OK == (res = _read_certificate(state, SLOTS[i], buf, &cbBuf))) {
+    if ((res == YKPIV_OK) && (cbBuf > 0)) {
       // add current slot to result, grow result buffer if necessary
 
       cbRealloc = (sizeof(ykpiv_key) + cbBuf - 1) > (cbData - offset) ? MAX((sizeof(ykpiv_key) + cbBuf - 1) - (cbData - offset), CB_PAGE) : 0;
 
       if (0 != cbRealloc) {
-        if (NULL == (pData = _ykpiv_realloc(state, pData, cbData + cbRealloc))) {
+        if (!(pTemp = _ykpiv_realloc(state, pData, cbData + cbRealloc))) {
+          /* realloc failed, pData will be freed in cleanup */
           res = YKPIV_MEMORY_ERROR;
           goto Cleanup;
         }
+        pData = pTemp;
+        pTemp = NULL;
       }
 
       cbData += cbRealloc;
@@ -337,14 +368,22 @@ ykpiv_rc ykpiv_util_read_cert(ykpiv_state *state, uint8_t slot, uint8_t **data,
 
   if ((NULL == data )|| (NULL == data_len)) return YKPIV_GENERIC_ERROR;
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   *data = 0;
   *data_len = 0;
 
   if (YKPIV_OK == (res = _read_certificate(state, slot, buf, &cbBuf))) {
-    if (NULL == (*data = _ykpiv_alloc(state, cbBuf))) {
+
+    /* handle those who write empty certificate blobs to PIV objects */
+    if (cbBuf == 0) {
+      *data = NULL;
+      *data_len = 0;
+      goto Cleanup;
+    }
+
+    if (!(*data = _ykpiv_alloc(state, cbBuf))) {
       res = YKPIV_MEMORY_ERROR;
       goto Cleanup;
     }
@@ -363,7 +402,7 @@ Cleanup:
 ykpiv_rc ykpiv_util_write_cert(ykpiv_state *state, uint8_t slot, uint8_t *data, size_t data_len, uint8_t certinfo) {
   ykpiv_rc res = YKPIV_OK;
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   res = _write_certificate(state, slot, data, data_len, certinfo);
@@ -390,7 +429,7 @@ ykpiv_rc ykpiv_util_block_puk(ykpiv_state *state) {
 
   if (NULL == state) return YKPIV_GENERIC_ERROR;
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   while (tries != 0) {
@@ -416,7 +455,7 @@ ykpiv_rc ykpiv_util_block_puk(ykpiv_state *state) {
         memcpy(&flags, p_item, cb_item);
       }
       else {
-        if (state->verbose) { fprintf(stderr, "admin flags exist, but are incorrect size = %zu", cb_item); }
+        if (state->verbose) { fprintf(stderr, "admin flags exist, but are incorrect size = %lu", (unsigned long)cb_item); }
       }
     }
   }
@@ -441,31 +480,31 @@ Cleanup:
 ykpiv_rc ykpiv_util_read_mscmap(ykpiv_state *state, ykpiv_container **containers, size_t *n_containers) {
   ykpiv_rc res = YKPIV_OK;
   uint8_t buf[CB_BUF_MAX];
-  size_t cbBuf = sizeof(buf);
+  unsigned long cbBuf = sizeof(buf);
   size_t len = 0;
   uint8_t *ptr = NULL;
 
   if ((NULL == containers) || (NULL == n_containers)) { res = YKPIV_GENERIC_ERROR; goto Cleanup; }
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   *containers = 0;
   *n_containers = 0;
 
-  if (YKPIV_OK == (res = _ykpiv_fetch_object(state, YKPIV_OBJ_MSCMAP, buf, (unsigned long*)&cbBuf))) {
+  if (YKPIV_OK == (res = _ykpiv_fetch_object(state, YKPIV_OBJ_MSCMAP, buf, &cbBuf))) {
     ptr = buf;
 
-    // check that object contents are at least large enough to read the header
+    /* check that object contents are at least large enough to read the header */
     if (cbBuf < CB_OBJ_TAG_MIN) {
       res = YKPIV_OK;
       goto Cleanup;
     }
 
     if (*ptr++ == TAG_MSCMAP) {
-      ptr += _ykpiv_get_length(ptr, &len);
+      ptr += (unsigned long)_ykpiv_get_length(ptr, &len);
 
-      // check that decoded length represents object contents
-      if (len > (cbBuf - (ptr - buf))) {
+      /* check that decoded length represents object contents */
+      if (len > (cbBuf - (size_t)(ptr - buf))) {
         res = YKPIV_OK;
         goto Cleanup;
       }
@@ -475,7 +514,7 @@ ykpiv_rc ykpiv_util_read_mscmap(ykpiv_state *state, ykpiv_container **containers
         goto Cleanup;
       }
 
-      // should check if container map isn't corrupt
+      /* should check if container map isn't corrupt */
 
       memcpy(*containers, ptr, len);
       *n_containers = len / sizeof(ykpiv_container);
@@ -491,12 +530,11 @@ Cleanup:
 ykpiv_rc ykpiv_util_write_mscmap(ykpiv_state *state, ykpiv_container *containers, size_t n_containers) {
   ykpiv_rc res = YKPIV_OK;
   uint8_t buf[CB_OBJ_MAX];
-  size_t cbBuf = sizeof(buf);
   size_t offset = 0;
   size_t req_len = 0;
   size_t data_len = n_containers * sizeof(ykpiv_container);
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   // check if data and data_len are zero, this means that
@@ -518,7 +556,7 @@ ykpiv_rc ykpiv_util_write_mscmap(ykpiv_state *state, ykpiv_container *containers
   // encode object data for storage
 
   // calculate the required length of the encoded object
-  req_len = 1 /* data tag */ + _ykpiv_set_length(buf, data_len) + data_len;
+  req_len = 1 /* data tag */ + (unsigned long)_ykpiv_set_length(buf, data_len) + data_len;
 
   if (req_len > _obj_size_max(state)) {
     res = YKPIV_SIZE_ERROR;
@@ -542,19 +580,20 @@ Cleanup:
 ykpiv_rc ykpiv_util_read_msroots(ykpiv_state *state, uint8_t **data, size_t *data_len) {
   ykpiv_rc res = YKPIV_OK;
   uint8_t buf[CB_BUF_MAX];
-  size_t cbBuf = sizeof(buf);
+  unsigned long cbBuf = sizeof(buf);
   size_t len = 0;
   uint8_t *ptr = NULL;
   int object_id = 0;
   uint8_t tag = 0;
   uint8_t *pData = NULL;
+  uint8_t *pTemp = NULL;
   size_t cbData = 0;
   size_t cbRealloc = 0;
   size_t offset = 0;
 
   if (!data || !data_len) return YKPIV_GENERIC_ERROR;
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   *data = 0;
@@ -567,7 +606,7 @@ ykpiv_rc ykpiv_util_read_msroots(ykpiv_state *state, uint8_t **data, size_t *dat
   for (object_id = YKPIV_OBJ_MSROOTS1; object_id <= YKPIV_OBJ_MSROOTS5; object_id++) {
     cbBuf = sizeof(buf);
 
-    if (YKPIV_OK != (res = _ykpiv_fetch_object(state, object_id, buf, (unsigned long*)&cbBuf))) {
+    if (YKPIV_OK != (res = _ykpiv_fetch_object(state, object_id, buf, &cbBuf))) {
       goto Cleanup;
     }
 
@@ -590,7 +629,7 @@ ykpiv_rc ykpiv_util_read_msroots(ykpiv_state *state, uint8_t **data, size_t *dat
     ptr += _ykpiv_get_length(ptr, &len);
 
     // check that decoded length represents object contents
-    if (len > (cbBuf - (ptr - buf))) {
+    if (len > (cbBuf - (size_t)(ptr - buf))) {
       res = YKPIV_OK;
       goto Cleanup;
     }
@@ -598,10 +637,13 @@ ykpiv_rc ykpiv_util_read_msroots(ykpiv_state *state, uint8_t **data, size_t *dat
     cbRealloc = len > (cbData - offset) ? len - (cbData - offset) : 0;
 
     if (0 != cbRealloc) {
-      if (NULL == (pData = _ykpiv_realloc(state, pData, cbData + cbRealloc))) {
+      if (!(pTemp = _ykpiv_realloc(state, pData, cbData + cbRealloc))) {
+        /* realloc failed, pData will be freed in cleanup */
         res = YKPIV_MEMORY_ERROR;
         goto Cleanup;
       }
+      pData = pTemp;
+      pTemp = NULL;
     }
 
     cbData += cbRealloc;
@@ -639,7 +681,7 @@ ykpiv_rc ykpiv_util_write_msroots(ykpiv_state *state, uint8_t *data, size_t data
   unsigned int i = 0;
   size_t cb_obj_max = _obj_size_max(state);
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   // check if either data and data_len are zero, this means that
@@ -673,14 +715,14 @@ ykpiv_rc ykpiv_util_write_msroots(ykpiv_state *state, uint8_t *data, size_t data
     offset = 0;
     data_chunk = MIN(cb_obj_max - CB_OBJ_TAG_MAX, data_len - data_offset);
 
-    // encode object data for storage
+    /* encode object data for storage */
     buf[offset++] = (i == (n_objs - 1)) ? TAG_MSROOTS_END : TAG_MSROOTS_MID;
     offset += _ykpiv_set_length(buf + offset, data_chunk);
     memcpy(buf + offset, data + data_offset, data_chunk);
     offset += data_chunk;
 
-    // write onto device
-    res = _ykpiv_save_object(state, YKPIV_OBJ_MSROOTS1 + i, buf, offset);
+    /* write onto device */
+    res = _ykpiv_save_object(state, (int)(YKPIV_OBJ_MSROOTS1 + i), buf, offset);
 
     if (YKPIV_OK != res) {
       goto Cleanup;
@@ -699,7 +741,6 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
   ykpiv_rc res = YKPIV_OK;
   unsigned char in_data[11];
   unsigned char *in_ptr = in_data;
-  char version[7];
   unsigned char data[1024];
   unsigned char templ[] = { 0, YKPIV_INS_GENERATE_ASYMMETRIC, 0, 0 };
   unsigned long recv_len = sizeof(data);
@@ -711,21 +752,47 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
   uint8_t *ptr_point = NULL;
   size_t  cb_point = 0;
 
+  setting_bool_t setting_roca = { 0 };
+  const char sz_setting_roca[] = "Enable_Unsafe_Keygen_ROCA";
+  const char sz_roca_format[] = "YubiKey serial number %u is affected by vulnerability "
+    "CVE-2017-15361 (ROCA) and should be replaced. On-chip key generation %s  "
+    "See YSA-2017-01 <https://www.yubico.com/support/security-advisories/ysa-2017-01/> "
+    "for additional information on device replacement and mitigation assistance.\n";
+  const char sz_roca_allow_user[] = "was permitted by an end-user configuration setting, but is not recommended.";
+  const char sz_roca_allow_admin[] = "was permitted by an administrator configuration setting, but is not recommended.";
+  const char sz_roca_block_user[] = "was blocked due to an end-user configuration setting.";
+  const char sz_roca_block_admin[] = "was blocked due to an administrator configuration setting.";
+  const char sz_roca_default[] = "was permitted by default, but is not recommended.  "
+    "The default behavior will change in a future Yubico release.";
+
+  if (!state) return YKPIV_ARGUMENT_ERROR;
+
   if (ykpiv_util_devicemodel(state) == DEVTYPE_YK4 && (algorithm == YKPIV_ALGO_RSA1024 || algorithm == YKPIV_ALGO_RSA2048)) {
-    if ((res = ykpiv_get_version(state, version, sizeof(version))) == YKPIV_OK) {
-      int major, minor, build;
-#pragma warning(push)
-#pragma warning(suppress: 4996) // Suppress _CRT_SECURE_NO_WARNINGS complaint (about sscanf_s)
-      int match = sscanf(version, "%d.%d.%d", &major, &minor, &build);
-#pragma warning(pop)
-      if (match == 3 && major == 4 && (minor < 3 || (minor == 3 && build < 5))) {
-        fprintf(stderr, "On-chip RSA key generation on this YubiKey has been blocked.\n");
-        fprintf(stderr, "Please see https://yubi.co/ysa201701/ for details.\n");
+    if ((state->ver.major == 4) && (state->ver.minor < 3 || ((state->ver.minor == 3) && (state->ver.patch < 5)))) {
+      const char *psz_msg = NULL;
+      setting_roca = setting_get_bool(sz_setting_roca, true);
+
+      switch (setting_roca.source) {
+        case SETTING_SOURCE_ADMIN:
+          psz_msg = setting_roca.value ? sz_roca_allow_admin : sz_roca_block_admin;
+          break;
+
+        case SETTING_SOURCE_USER:
+          psz_msg = setting_roca.value ? sz_roca_allow_user : sz_roca_block_user;
+          break;
+
+        default:
+        case SETTING_SOURCE_DEFAULT:
+          psz_msg = sz_roca_default;
+          break;
+      }
+
+      fprintf(stderr, sz_roca_format, state->serial, psz_msg);
+      yc_log_event(1, setting_roca.value ? YC_LOG_LEVEL_WARN : YC_LOG_LEVEL_ERROR, sz_roca_format, state->serial, psz_msg);
+
+      if (!setting_roca.value) {
         return YKPIV_NOT_SUPPORTED;
       }
-    } else {
-      fprintf(stderr, "Failed to get device version.\n");
-      return YKPIV_GENERIC_ERROR;
     }
   }
 
@@ -757,7 +824,7 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
     return YKPIV_GENERIC_ERROR;
   }
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   templ[3] = slot;
@@ -788,7 +855,7 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
     *in_ptr++ = touch_policy;
   }
 
-  if (YKPIV_OK != (res = ykpiv_transfer_data(state, templ, in_data, (long)(in_ptr - in_data), data, &recv_len, &sw))) {
+  if (YKPIV_OK != (res = _ykpiv_transfer_data(state, templ, in_data, (long)(in_ptr - in_data), data, &recv_len, &sw))) {
     if (state->verbose) { fprintf(stderr, "Failed to communicate.\n"); }
     goto Cleanup;
   }
@@ -949,7 +1016,7 @@ ykpiv_rc ykpiv_util_get_config(ykpiv_state *state, ykpiv_config *config) {
   config->pin_last_changed = 0;
   config->mgm_type = YKPIV_CONFIG_MGM_MANUAL;
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   /* recover admin data */
@@ -1017,7 +1084,7 @@ ykpiv_rc ykpiv_util_set_pin_last_changed(ykpiv_state *state) {
 
   if (NULL == state) return YKPIV_GENERIC_ERROR;
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   /* recover admin data */
@@ -1054,14 +1121,14 @@ ykpiv_rc ykpiv_util_get_derived_mgm(ykpiv_state *state, const uint8_t *pin, cons
   if (NULL == state) return YKPIV_GENERIC_ERROR;
   if ((NULL == pin) || (0 == pin_len) || (NULL == mgm)) return YKPIV_GENERIC_ERROR;
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   /* recover management key */
   if (YKPIV_OK == (res = _read_metadata(state, TAG_ADMIN, data, &cb_data))) {
     if (YKPIV_OK == (res = _get_metadata_item(data, cb_data, TAG_ADMIN_SALT, &p_item, &cb_item))) {
       if (cb_item != CB_ADMIN_SALT) {
-        if (state->verbose) fprintf(stderr, "derived mgm salt exists, but is incorrect size = %zu\n", cb_item);
+        if (state->verbose) fprintf(stderr, "derived mgm salt exists, but is incorrect size = %lu\n", (unsigned long)cb_item);
         res = YKPIV_GENERIC_ERROR;
         goto Cleanup;
       }
@@ -1090,7 +1157,7 @@ ykpiv_rc ykpiv_util_get_protected_mgm(ykpiv_state *state, ykpiv_mgm *mgm) {
   if (NULL == state) return YKPIV_GENERIC_ERROR;
   if (NULL == mgm) return YKPIV_GENERIC_ERROR;
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   if (YKPIV_OK != (res = _read_metadata(state, TAG_PROTECTED, data, &cb_data))) {
@@ -1104,7 +1171,7 @@ ykpiv_rc ykpiv_util_get_protected_mgm(ykpiv_state *state, ykpiv_mgm *mgm) {
   }
 
   if (cb_item != member_size(ykpiv_mgm, data)) {
-    if (state->verbose) fprintf(stderr, "protected data contains mgm, but is the wrong size = %zu\n", cb_item);
+    if (state->verbose) fprintf(stderr, "protected data contains mgm, but is the wrong size = %lu\n", (unsigned long)cb_item);
     res = YKPIV_AUTHENTICATION_ERROR;
     goto Cleanup;
   }
@@ -1113,7 +1180,7 @@ ykpiv_rc ykpiv_util_get_protected_mgm(ykpiv_state *state, ykpiv_mgm *mgm) {
 
 Cleanup:
 
-  memset(data, 0, sizeof(data));
+  yc_memzero(data, sizeof(data));
 
   _ykpiv_end_transaction(state);
   return res;
@@ -1151,7 +1218,7 @@ ykpiv_rc ykpiv_util_set_protected_mgm(ykpiv_state *state, ykpiv_mgm *mgm) {
     }
   }
 
-  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return YKPIV_PCSC_ERROR;
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) goto Cleanup;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   /* try to set the mgm key as long as we don't encounter a fatal error */
@@ -1222,7 +1289,7 @@ ykpiv_rc ykpiv_util_set_protected_mgm(ykpiv_state *state, ykpiv_mgm *mgm) {
       memcpy(&flags_1, p_item, cb_item);
     }
     else {
-      if (state->verbose) fprintf(stderr, "admin data flags are an incorrect size = %zu\n", cb_item);
+      if (state->verbose) fprintf(stderr, "admin data flags are an incorrect size = %lu\n", (unsigned long)cb_item);
     }
 
     /* remove any existing salt */
@@ -1246,8 +1313,8 @@ ykpiv_rc ykpiv_util_set_protected_mgm(ykpiv_state *state, ykpiv_mgm *mgm) {
 
 Cleanup:
 
-  memset(data, 0, sizeof(data));
-  memset(mgm_key, 0, sizeof(mgm_key));
+  yc_memzero(data, sizeof(data));
+  yc_memzero(mgm_key, sizeof(mgm_key));
 
   _ykpiv_end_transaction(state);
   return res;
@@ -1280,7 +1347,7 @@ uint32_t ykpiv_util_slot_object(uint8_t slot) {
     object_id = YKPIV_OBJ_SIGNATURE;
     break;
 
-  case  YKPIV_KEY_KEYMGM:
+  case YKPIV_KEY_KEYMGM:
     object_id = YKPIV_OBJ_KEY_MANAGEMENT;
     break;
 
@@ -1299,13 +1366,13 @@ uint32_t ykpiv_util_slot_object(uint8_t slot) {
     break;
   }
 
-  return object_id;
+  return (uint32_t)object_id;
 }
 
 static ykpiv_rc _read_certificate(ykpiv_state *state, uint8_t slot, uint8_t *buf, size_t *buf_len) {
   ykpiv_rc res = YKPIV_OK;
   uint8_t *ptr = NULL;
-  int object_id = ykpiv_util_slot_object(slot);
+  int object_id = (int)ykpiv_util_slot_object(slot);
   size_t len = 0;
 
   if (-1 == object_id) return YKPIV_INVALID_OBJECT;
@@ -1325,7 +1392,7 @@ static ykpiv_rc _read_certificate(ykpiv_state *state, uint8_t slot, uint8_t *buf
       ptr += _ykpiv_get_length(ptr, &len);
 
       // check that decoded length represents object contents
-      if (len > (*buf_len - (ptr - buf))) {
+      if (len > (*buf_len - (size_t)(ptr - buf))) {
         *buf_len = 0;
         return YKPIV_OK;
       }
@@ -1343,8 +1410,7 @@ static ykpiv_rc _read_certificate(ykpiv_state *state, uint8_t slot, uint8_t *buf
 
 static ykpiv_rc _write_certificate(ykpiv_state *state, uint8_t slot, uint8_t *data, size_t data_len, uint8_t certinfo) {
   uint8_t buf[CB_OBJ_MAX];
-  size_t cbBuf = sizeof(buf);
-  int object_id = ykpiv_util_slot_object(slot);
+  int object_id = (int)ykpiv_util_slot_object(slot);
   size_t offset = 0;
   size_t req_len = 0;
 
@@ -1367,8 +1433,10 @@ static ykpiv_rc _write_certificate(ykpiv_state *state, uint8_t slot, uint8_t *da
   // calculate the required length of the encoded object
   req_len = 1 /* cert tag */ + 3 /* compression tag + data*/ + 2 /* lrc */;
   req_len += _ykpiv_set_length(buf, data_len);
+  req_len += data_len;
 
-  if (req_len > _obj_size_max(state)) return YKPIV_SIZE_ERROR;
+  if (req_len < data_len) return YKPIV_SIZE_ERROR; /* detect overflow of unsigned size_t */
+  if (req_len > _obj_size_max(state)) return YKPIV_SIZE_ERROR; /* obj_size_max includes limits for TLV encoding */
 
   buf[offset++] = TAG_CERT;
   offset += _ykpiv_set_length(buf + offset, data_len);
@@ -1415,6 +1483,11 @@ static ykpiv_rc _get_metadata_item(uint8_t *data, size_t cb_data, uint8_t tag, u
 
   while (p_temp < (data + cb_data)) {
     tag_temp = *p_temp++;
+
+    if (!_ykpiv_has_valid_length(p_temp, (size_t)(data + cb_data - p_temp))) {
+      return YKPIV_SIZE_ERROR;
+    }
+
     p_temp += _ykpiv_get_length(p_temp, &cb_temp);
 
     if (tag_temp == tag) {
@@ -1425,7 +1498,8 @@ static ykpiv_rc _get_metadata_item(uint8_t *data, size_t cb_data, uint8_t tag, u
     p_temp += cb_temp;
   }
 
-  if (p_temp < (data + cb_data)) {
+  // Make sure the item doesn't end after the buffer
+  if ((p_temp + cb_temp) <= (data + cb_data)) {
     *pp_item = p_temp;
     *pcb_item = cb_temp;
     return YKPIV_OK;
@@ -1434,6 +1508,44 @@ static ykpiv_rc _get_metadata_item(uint8_t *data, size_t cb_data, uint8_t tag, u
   return YKPIV_GENERIC_ERROR;
 }
 
+ykpiv_rc ykpiv_util_parse_metadata(uint8_t *data, size_t data_len, ykpiv_metadata *metadata) {
+  uint8_t *p;
+  size_t cb;
+
+  ykpiv_rc rc = _get_metadata_item(data, data_len, YKPIV_METADATA_ALGORITHM_TAG, &p, &cb);
+  if(rc != YKPIV_OK)
+    return rc;
+  if(cb != 1)
+    return YKPIV_PARSE_ERROR;
+  metadata->algorithm = p[0];
+
+  rc = _get_metadata_item(data, data_len, YKPIV_METADATA_POLICY_TAG, &p, &cb);
+  if(rc != YKPIV_OK)
+    return rc;
+  if(cb != 2)
+    return YKPIV_PARSE_ERROR;
+  metadata->pin_policy = p[0];
+  metadata->touch_policy = p[1];
+
+  rc = _get_metadata_item(data, data_len, YKPIV_METADATA_ORIGIN_TAG, &p, &cb);
+  if(rc != YKPIV_OK)
+    return rc;
+  if(cb != 1)
+    return YKPIV_PARSE_ERROR;
+  metadata->origin = p[0];
+
+  rc = _get_metadata_item(data, data_len, YKPIV_METADATA_PUBKEY_TAG, &p, &cb);
+  if(rc != YKPIV_OK)
+    return rc;
+  if(cb > sizeof(metadata->pubkey))
+    return YKPIV_PARSE_ERROR;
+
+  metadata->pubkey_len = cb;
+  memcpy(metadata->pubkey, p, cb);
+
+  return YKPIV_OK;
+}
+
 static int _get_length_size(size_t length) {
   if (length < 0x80) {
     return 1;
@@ -1482,16 +1594,24 @@ static ykpiv_rc _set_metadata_item(uint8_t *data, size_t *pcb_data, size_t cb_da
 
       /* length doesn't match, expand/shrink to fit */
       p_next = p_temp + cb_temp;
-      cb_moved = (long)cb_item - (long)cb_temp + ((long)(cb_item != 0 ? _get_length_size(cb_item) : -1 /* for tag, if deleting */) - (long)cb_len); /* accounts for different length encoding */
+      cb_moved = (long)cb_item - (long)cb_temp +
+        ((long)(cb_item != 0 ? _get_length_size(cb_item) : -1 /* for tag, if deleting */) -
+        (long)cb_len); /* accounts for different length encoding */
 
       /* length would cause buffer overflow, return error */
-      if (*pcb_data + cb_moved > cb_data_max) {
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wsign-conversion"
+      if ((size_t)(*pcb_data + cb_moved) > cb_data_max) {
         return YKPIV_GENERIC_ERROR;
       }
+#pragma GCC diagnostic pop
 
       /* move remaining data */
-      memmove(p_next + cb_moved, p_next, *pcb_data - (p_next - data));
+      memmove(p_next + cb_moved, p_next, *pcb_data - (size_t)(p_next - data));
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wsign-conversion"
       *pcb_data += cb_moved;
+#pragma GCC diagnostic pop
 
       /* re-encode item and insert */
       if (cb_item != 0) {
@@ -1501,7 +1621,7 @@ static ykpiv_rc _set_metadata_item(uint8_t *data, size_t *pcb_data, size_t cb_da
       }
 
       return YKPIV_OK;
-    } //if tag found
+    } /* if tag found */
 
     p_temp += cb_temp;
   }
@@ -1513,7 +1633,7 @@ static ykpiv_rc _set_metadata_item(uint8_t *data, size_t *pcb_data, size_t cb_da
 
   // we did not find an existing tag, append
   p_temp = data + *pcb_data;
-  cb_len = _get_length_size(cb_item);
+  cb_len = (size_t)_get_length_size(cb_item);
 
   // length would cause buffer overflow, return error
   if (*pcb_data + cb_len + cb_item > cb_data_max) {
@@ -1543,7 +1663,7 @@ static ykpiv_rc _set_metadata_item(uint8_t *data, size_t *pcb_data, size_t cb_da
 static ykpiv_rc _read_metadata(ykpiv_state *state, uint8_t tag, uint8_t* data, size_t* pcb_data) {
   ykpiv_rc res = YKPIV_OK;
   uint8_t *p_temp = NULL;
-  size_t cb_temp = 0;
+  unsigned long cb_temp = 0;
   int obj_id = 0;
 
   if (!data || !pcb_data || (CB_BUF_MAX > *pcb_data)) return YKPIV_GENERIC_ERROR;
@@ -1557,7 +1677,7 @@ static ykpiv_rc _read_metadata(ykpiv_state *state, uint8_t tag, uint8_t* data, s
   cb_temp = *pcb_data;
   *pcb_data = 0;
 
-  if (YKPIV_OK != (res = _ykpiv_fetch_object(state, obj_id, data, (unsigned long*)&cb_temp))) {
+  if (YKPIV_OK != (res = _ykpiv_fetch_object(state, obj_id, data, &cb_temp))) {
     return res;
   }
 
@@ -1569,7 +1689,7 @@ static ykpiv_rc _read_metadata(ykpiv_state *state, uint8_t tag, uint8_t* data, s
 
   p_temp += _ykpiv_get_length(p_temp, pcb_data);
 
-  if (*pcb_data > ((size_t)cb_temp - (p_temp - data))) {
+  if (*pcb_data > (cb_temp - (size_t)(p_temp - data))) {
     *pcb_data = 0;
     return YKPIV_GENERIC_ERROR;
   }
@@ -1615,7 +1735,7 @@ static ykpiv_rc _write_metadata(ykpiv_state *state, uint8_t tag, uint8_t *data,
     memcpy(pTemp, data, cb_data);
     pTemp += cb_data;
 
-    res = _ykpiv_save_object(state, obj_id, buf, pTemp - buf);
+    res = _ykpiv_save_object(state, obj_id, buf, (size_t)(pTemp - buf));
   }
 
   return res;
