@@ -31,6 +31,10 @@ void RSA_get0_crt_params(const RSA *r,
                          const BIGNUM **iqmp);
 void X509_SIG_getm(X509_SIG *sig, X509_ALGOR **palg,
                    ASN1_OCTET_STRING **pdigest);
+int ECDSA_SIG_set0(ECDSA_SIG *sig, BIGNUM *r, BIGNUM *s);
+void ECDSA_SIG_get0(const ECDSA_SIG *sig, const BIGNUM **pr, const BIGNUM **ps);
+RSA *EVP_PKEY_get0_RSA(const EVP_PKEY *pkey);
+EC_KEY *EVP_PKEY_get0_EC_KEY(const EVP_PKEY *pkey);
 
 #endif /* _WINDOWS */
 #endif /* OPENSSL_VERSION_NUMBER || LIBRESSL_VERSION_NUMBER */
