@@ -43,16 +43,18 @@
 #endif
 
 #include "openssl-compat.h"
+#include <openssl/bn.h>
 #include <openssl/des.h>
 #include <openssl/pem.h>
 #include <openssl/pkcs12.h>
 #include <openssl/rand.h>
+#include <openssl/rsa.h>
+#include <openssl/x509v3.h>
 
 #include "cmdline.h"
 #include "util.h"
 
 #define MAX(a,b) (a) > (b) ? (a) : (b)
-#define MIN(a,b) (a) < (b) ? (a) : (b)
 
 #define CHUID 0
 #define CCC 1
@@ -124,48 +126,68 @@ static bool sign_data(ykpiv_state *state, const unsigned char *in, size_t len, u
   return false;
 }
 
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if !((OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER))
 static int ec_key_ex_data_idx = -1;
 
 struct internal_key {
   ykpiv_state *state;
   int algorithm;
   int key;
+  const unsigned char *oid;
+  size_t oid_len;
 };
 
 int yk_rsa_meth_sign(int dtype, const unsigned char *m, unsigned int m_length,
     unsigned char *sigret, unsigned int *siglen, const RSA *rsa) {
+  size_t yk_siglen = RSA_size(rsa);
   const RSA_METHOD *meth = RSA_get_method(rsa);
   const struct internal_key *key = RSA_meth_get0_app_data(meth);
-  if (sign_data(key->state, m, m_length, sigret, (size_t *)siglen, key->algorithm, key->key))
-    return 0;
+  unsigned char message[256] = {0};
 
-  return 1;
+  if(key->oid_len) {
+    memcpy(message, key->oid, key->oid_len);
+    memcpy(message + key->oid_len, m, m_length);
+    m_length += key->oid_len;
+    m = message;
+  }
+  if (sign_data(key->state, m, m_length, sigret, &yk_siglen, key->algorithm, key->key)) {
+    *siglen = (unsigned int)yk_siglen;
+    return 1;
+  }
+
+  return 0;
 }
 
 int yk_ec_meth_sign(int type, const unsigned char *dgst, int dlen,
     unsigned char *sig, unsigned int *siglen, const BIGNUM *kinv,
     const BIGNUM *r, EC_KEY *ec) {
+  size_t yk_siglen = ECDSA_size(ec);
   const struct internal_key *key = EC_KEY_get_ex_data(ec, ec_key_ex_data_idx);
-  if (sign_data(key->state, dgst, dlen, sig, (size_t *)siglen, key->algorithm, key->key))
-    return 0;
+  if (sign_data(key->state, dgst, dlen, sig, &yk_siglen, key->algorithm, key->key)) {
+    *siglen = (unsigned int)yk_siglen;
+    return 1;
+  }
 
-  return 1;
+  return 0;
 }
 
 static int wrap_public_key(ykpiv_state *state, int algorithm, EVP_PKEY *public_key,
-    int key) {
+    int key, const unsigned char *oid, size_t oid_len) {
+  static struct internal_key int_key;
+  int_key.state = state;
+  int_key.algorithm = algorithm;
+  int_key.key = key;
+  int_key.oid = oid;
+  int_key.oid_len = oid_len;
   if(YKPIV_IS_RSA(algorithm)) {
     RSA_METHOD *meth = RSA_meth_dup(RSA_get_default_method());
     RSA *rsa = EVP_PKEY_get0_RSA(public_key);
-    struct internal_key int_key = {state, algorithm, key};
     RSA_meth_set0_app_data(meth, &int_key);
     RSA_meth_set_sign(meth, yk_rsa_meth_sign);
     RSA_set_method(rsa, meth);
   } else {
     EC_KEY *ec = EVP_PKEY_get0_EC_KEY(public_key);
     EC_KEY_METHOD *meth = EC_KEY_METHOD_new(EC_KEY_get_method(ec));
-    struct internal_key int_key = {state, algorithm, key};
     if (ec_key_ex_data_idx == -1)
       ec_key_ex_data_idx = EC_KEY_get_ex_new_index(0, NULL, NULL, NULL, 0);
     EC_KEY_set_ex_data(ec, ec_key_ex_data_idx, &int_key);
@@ -428,11 +450,11 @@ static bool import_key(ykpiv_state *state, enum enum_key_format key_format,
       }
 
       rc = ykpiv_import_private_key(state, key, algorithm,
-                                    p, element_len,
-                                    q, element_len,
-                                    dmp1, element_len,
-                                    dmq1, element_len,
-                                    iqmp, element_len,
+                                    p, (size_t)element_len,
+                                    q, (size_t)element_len,
+                                    dmp1, (size_t)element_len,
+                                    dmq1, (size_t)element_len,
+                                    iqmp, (size_t)element_len,
                                     NULL, 0,
                                     pp, tp);
     }
@@ -550,6 +572,12 @@ static bool import_cert(ykpiv_state *state, enum enum_key_format cert_format,
     unsigned char certdata[YKPIV_OBJ_MAX_SIZE];
     unsigned char *certptr = certdata;
     ykpiv_rc res;
+
+    if(cert_len > YKPIV_OBJ_MAX_SIZE) {
+      fprintf(stderr, "Length of certificate is more than can fit.\n");
+      goto import_cert_out;
+    }
+
     if (compress) {
       if (fread(certdata, 1, (size_t)cert_len, input_file) != (size_t)cert_len) {
         fprintf(stderr, "Failed to read compressed certificate\n");
@@ -558,7 +586,7 @@ static bool import_cert(ykpiv_state *state, enum enum_key_format cert_format,
     } else {
       i2d_X509(cert, &certptr);
     }
-    if ((res = ykpiv_util_write_cert(state, get_slot_hex(slot), certdata, cert_len, compress)) != YKPIV_OK) {
+    if ((res = ykpiv_util_write_cert(state, get_slot_hex(slot), certdata, (size_t)cert_len, compress)) != YKPIV_OK) {
       fprintf(stderr, "Failed commands with device: %s\n", ykpiv_strerror(res));
     } else {
       ret = true;
@@ -616,20 +644,19 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
   EVP_PKEY *public_key = NULL;
   const EVP_MD *md;
   bool ret = false;
-  unsigned char digest[EVP_MAX_MD_SIZE + MAX_OID_LEN];
-  unsigned int digest_len;
-  unsigned int md_len;
   unsigned char algorithm;
   int key = 0;
-  unsigned char *signinput;
-  size_t len = 0;
   size_t oid_len;
   const unsigned char *oid;
+#if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
+  unsigned char digest[EVP_MAX_MD_SIZE + MAX_OID_LEN];
+  unsigned int md_len;
+  unsigned int digest_len;
+  unsigned char *signinput;
+  size_t len = 0;
   int nid;
   ASN1_TYPE null_parameter;
-
-  null_parameter.type = V_ASN1_NULL;
-  null_parameter.value.ptr = NULL;
+#endif
 
   key = get_slot_hex(slot);
 
@@ -663,9 +690,6 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
     goto request_out;
   }
 
-  md_len = (unsigned int)EVP_MD_size(md);
-  digest_len = sizeof(digest) - md_len;
-
   req = X509_REQ_new();
   if(!req) {
     fprintf(stderr, "Failed to allocate request structure.\n");
@@ -688,7 +712,13 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
     goto request_out;
   }
 
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
+  null_parameter.type = V_ASN1_NULL;
+  null_parameter.value.ptr = NULL;
+
+  md_len = (unsigned int)EVP_MD_size(md);
+  digest_len = sizeof(digest) - md_len;
+
   memcpy(digest, oid, oid_len);
   /* XXX: this should probably use X509_REQ_digest() but that's buggy */
   if(!ASN1_item_digest(ASN1_ITEM_rptr(X509_REQ_INFO), md, req->req_info,
@@ -721,16 +751,19 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
       fprintf(stderr, "Failed signing request.\n");
       goto request_out;
     }
-    M_ASN1_BIT_STRING_set(req->signature, signature, sig_len);
+    ASN1_STRING_set(req->signature, signature, sig_len);
     /* mark that all bits should be used. */
     req->signature->flags = ASN1_STRING_FLAG_BITS_LEFT;
   }
 #else
   /* With opaque structures we can not touch whatever we want, but we need
    * to embed the sign_data function in the RSA/EC key structures  */
-  wrap_public_key(state, algorithm, public_key, key);
+  wrap_public_key(state, algorithm, public_key, key, oid, oid_len);
 
-  X509_REQ_sign(req, public_key, md);
+  if(X509_REQ_sign(req, public_key, md) == 0) {
+    fprintf(stderr, "Failed signing request.\n");
+    goto request_out;
+  }
 #endif
 
   if(key_format == key_format_arg_PEM) {
@@ -751,7 +784,7 @@ request_out:
     EVP_PKEY_free(public_key);
   }
   if(req) {
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
     if(req->sig_alg->parameter) {
       req->sig_alg->parameter = NULL;
     }
@@ -764,6 +797,16 @@ request_out:
   return ret;
 }
 
+static const struct {
+  int nid;
+  const char *ext;
+  int critical;
+} selfsign_extensions[] = {
+  {NID_subject_key_identifier, "hash", 0},
+  {NID_authority_key_identifier, "keyid", 0},
+  {NID_basic_constraints, "CA:true", 1},
+};
+
 static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_format,
     const char *input_file_name, enum enum_slot slot, char *subject, enum enum_hash hash,
     const int *serial, int validDays, const char *output_file_name) {
@@ -774,22 +817,21 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
   X509 *x509 = NULL;
   X509_NAME *name = NULL;
   const EVP_MD *md;
-  unsigned char digest[EVP_MAX_MD_SIZE + MAX_OID_LEN];
-  unsigned int digest_len;
   unsigned char algorithm;
   int key = 0;
-  unsigned char *signinput;
-  size_t len = 0;
   size_t oid_len;
   const unsigned char *oid;
   int nid;
-  unsigned int md_len;
   ASN1_INTEGER *sno = ASN1_INTEGER_new();
   BIGNUM *ser = NULL;
+#if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
+  unsigned char digest[EVP_MAX_MD_SIZE + MAX_OID_LEN];
+  unsigned int digest_len;
+  unsigned int md_len;
+  unsigned char *signinput;
+  size_t len = 0;
   ASN1_TYPE null_parameter;
-
-  null_parameter.type = V_ASN1_NULL;
-  null_parameter.value.ptr = NULL;
+#endif
 
   key = get_slot_hex(slot);
 
@@ -822,9 +864,6 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
   if(md == NULL) {
     goto selfsign_out;
   }
-  md_len = (unsigned int)EVP_MD_size(md);
-  digest_len = sizeof(digest) - md_len;
-
   x509 = X509_new();
   if(!x509) {
     fprintf(stderr, "Failed to allocate certificate structure.\n");
@@ -859,11 +898,19 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
     fprintf(stderr, "Failed to set certificate serial.\n");
     goto selfsign_out;
   }
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
   if(!X509_gmtime_adj(X509_get_notBefore(x509), 0)) {
+#else
+  if(!X509_gmtime_adj(X509_getm_notBefore(x509), 0)) {
+#endif
     fprintf(stderr, "Failed to set certificate notBefore.\n");
     goto selfsign_out;
   }
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
   if(!X509_gmtime_adj(X509_get_notAfter(x509), 60L * 60L * 24L * validDays)) {
+#else
+  if(!X509_gmtime_adj(X509_getm_notAfter(x509), 60L * 60L * 24L * validDays)) {
+#endif
     fprintf(stderr, "Failed to set certificate notAfter.\n");
     goto selfsign_out;
   }
@@ -884,7 +931,55 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
   if(nid == 0) {
     goto selfsign_out;
   }
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
+
+  {
+    X509V3_CTX ctx;
+    int i;
+    X509V3_set_ctx(&ctx, x509, x509, NULL, NULL, 0);
+
+    for(i = 0; i < sizeof(selfsign_extensions) / sizeof(selfsign_extensions[0]); i++) {
+      X509_EXTENSION *ext = NULL;
+      void *ext_struc;
+      const X509V3_EXT_METHOD *method = X509V3_EXT_get_nid(selfsign_extensions[i].nid);
+
+      if(!method) {
+        fprintf(stderr, "Failed to get extension method for nid %d.\n", selfsign_extensions[i].nid);
+        goto selfsign_out;
+      }
+      if(method->v2i) {
+        STACK_OF(CONF_VALUE) *nval = X509V3_parse_list(selfsign_extensions[i].ext);
+        if(!nval) {
+          fprintf(stderr, "Failed parsing extension value for nid %d.\n", selfsign_extensions[i].nid);
+          goto selfsign_out;
+        }
+        ext_struc = method->v2i(method, &ctx, nval);
+      } else if(method->s2i) {
+        ext_struc = method->s2i(method, &ctx, selfsign_extensions[i].ext);
+      } else {
+        fprintf(stderr, "Unknown way to construct extension for nid %d.\n", selfsign_extensions[i].nid);
+        goto selfsign_out;
+      }
+
+      if(!ext_struc) {
+        fprintf(stderr, "Failed constructing extension value for nid %d.\n", selfsign_extensions[i].nid);
+        goto selfsign_out;
+      }
+
+      ext = X509V3_EXT_i2d(selfsign_extensions[i].nid, selfsign_extensions[i].critical, ext_struc);
+      if(!X509_add_ext(x509, ext, -1)) {
+        fprintf(stderr, "Failed adding extension %d (%d).\n", i, selfsign_extensions[i].nid);
+        goto selfsign_out;
+      }
+    }
+  }
+
+#if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
+  null_parameter.type = V_ASN1_NULL;
+  null_parameter.value.ptr = NULL;
+
+  md_len = (unsigned int)EVP_MD_size(md);
+  digest_len = sizeof(digest) - md_len;
+
   if(YKPIV_IS_RSA(algorithm)) {
     signinput = digest;
     len = oid_len + md_len;
@@ -912,7 +1007,7 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
       fprintf(stderr, "Failed signing certificate.\n");
       goto selfsign_out;
     }
-    M_ASN1_BIT_STRING_set(x509->signature, signature, sig_len);
+    ASN1_STRING_set(x509->signature, signature, sig_len);
     /* setting flags to ASN1_STRING_FLAG_BITS_LEFT here marks that no bits
      * should be subtracted from the bit string, thus making sure that the
      * certificate can be validated. */
@@ -921,9 +1016,12 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
 #else
   /* With opaque structures we can not touch whatever we want, but we need
    * to embed the sign_data function in the RSA/EC key structures  */
-  wrap_public_key(state, algorithm, public_key, key);
+  wrap_public_key(state, algorithm, public_key, key, oid, oid_len);
 
-  X509_sign(x509, public_key, md);
+  if(X509_sign(x509, public_key, md) == 0) {
+    fprintf(stderr, "Failed signing certificate.\n");
+    goto selfsign_out;
+  }
 #endif
 
   if(key_format == key_format_arg_PEM) {
@@ -941,7 +1039,7 @@ selfsign_out:
     fclose(output_file);
   }
   if(x509) {
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
     if(x509->sig_alg->parameter) {
       x509->sig_alg->parameter = NULL;
       x509->cert_info->signature->parameter = NULL;
@@ -1021,6 +1119,9 @@ static bool change_pin(ykpiv_state *state, enum enum_action action, const char *
     op = ykpiv_change_puk;
   }
   res = op(state, pin, pin_len, new_pin, new_len, &tries);
+
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wswitch-enum"
   switch (res) {
     case YKPIV_OK:
       return true;
@@ -1042,6 +1143,7 @@ static bool change_pin(ykpiv_state *state, enum enum_action action, const char *
       fprintf(stderr, "Failed changing/unblocking code, error: %s\n", ykpiv_strerror(res));
       return false;
   }
+#pragma GCC diagnostic pop
 }
 
 static bool delete_certificate(ykpiv_state *state, enum enum_slot slot) {
@@ -1077,11 +1179,6 @@ static bool read_certificate(ykpiv_state *state, enum enum_slot slot,
 
   if (key_format == key_format_arg_PEM ||
       key_format == key_format_arg_SSH) {
-    x509 = X509_new();
-    if (!x509) {
-      fprintf(stderr, "Failed allocating x509 structure.\n");
-      goto read_cert_out;
-    }
     x509 = d2i_X509(NULL, (const unsigned char**)&ptr, cert_len);
     if (!x509) {
       fprintf(stderr, "Failed parsing x509 information.\n");
@@ -1212,12 +1309,12 @@ out:
 
 static void print_cert_info(ykpiv_state *state, enum enum_slot slot, const EVP_MD *md,
     FILE *output) {
-  int object = ykpiv_util_slot_object(get_slot_hex(slot));
+  int object = (int)ykpiv_util_slot_object(get_slot_hex(slot));
   int slot_name;
-  unsigned char data[3072];
+  unsigned char data[YKPIV_OBJ_MAX_SIZE];
   const unsigned char *ptr = data;
   unsigned long len = sizeof(data);
-  int cert_len;
+  unsigned long cert_len;
   X509 *x509 = NULL;
   X509_NAME *subj;
   BIO *bio = NULL;
@@ -1232,14 +1329,9 @@ static void print_cert_info(ykpiv_state *state, enum enum_slot slot, const EVP_M
 
   if(*ptr++ == 0x70) {
     unsigned int md_len = sizeof(data);
-    ASN1_TIME *not_before, *not_after;
+    const ASN1_TIME *not_before, *not_after;
 
     ptr += get_length(ptr, &cert_len);
-    x509 = X509_new();
-    if(!x509) {
-      fprintf(output, "Allocation failure.\n");
-      return;
-    }
     x509 = d2i_X509(NULL, &ptr, cert_len);
     if(!x509) {
       fprintf(output, "Unknown data present.\n");
@@ -1290,13 +1382,21 @@ static void print_cert_info(ykpiv_state *state, enum enum_slot slot, const EVP_M
     dump_data(data, md_len, output, false, format_arg_hex);
 
     bio = BIO_new_fp(output, BIO_NOCLOSE | BIO_FP_TEXT);
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
     not_before = X509_get_notBefore(x509);
+#else
+    not_before = X509_get0_notBefore(x509);
+#endif
     if(not_before) {
       fprintf(output, "\tNot Before:\t");
       ASN1_TIME_print(bio, not_before);
       fprintf(output, "\n");
     }
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
     not_after = X509_get_notAfter(x509);
+#else
+    not_after = X509_get0_notAfter(x509);
+#endif
     if(not_after) {
       fprintf(output, "\tNot After:\t");
       ASN1_TIME_print(bio, not_after);
@@ -1319,10 +1419,12 @@ static bool status(ykpiv_state *state, enum enum_hash hash,
                    enum enum_slot slot,
                    const char *output_file_name) {
   const EVP_MD *md;
-  unsigned char buf[3072];
+  unsigned char buf[YKPIV_OBJ_MAX_SIZE];
   long unsigned len = sizeof(buf);
   int i;
+  uint32_t serial = 0;
   FILE *output_file = open_file(output_file_name, OUTPUT_TEXT);
+
   if(!output_file) {
     return false;
   }
@@ -1332,6 +1434,20 @@ static bool status(ykpiv_state *state, enum enum_hash hash,
     return false;
   }
 
+  fprintf(output_file, "Version:\t");
+  if (ykpiv_get_version(state, (char*)buf, (size_t)len) != YKPIV_OK) {
+    fprintf(output_file, "No data available\n");
+  } else {
+    fprintf(output_file, "%s\n", (char*)buf);
+  }
+
+  fprintf(output_file, "Serial Number:\t");
+  if (ykpiv_get_serial(state, &serial) != YKPIV_OK) {
+    fprintf(output_file, "No data available\n");
+  } else {
+    fprintf(output_file, "%d\n", serial);
+  }
+
   fprintf(output_file, "CHUID:\t");
   if(ykpiv_fetch_object(state, YKPIV_OBJ_CHUID, buf, &len) != YKPIV_OK) {
     fprintf(output_file, "No data available\n");
@@ -1407,7 +1523,7 @@ static bool test_signature(ykpiv_state *state, enum enum_slot slot,
   {
     unsigned char rand[128];
     EVP_MD_CTX *mdctx;
-    if(RAND_bytes(rand, 128) == -1) {
+    if(RAND_bytes(rand, sizeof(rand)) <= 0) {
       fprintf(stderr, "error: no randomness.\n");
       return false;
     }
@@ -1554,7 +1670,7 @@ static bool test_decipher(ykpiv_state *state, enum enum_slot slot,
       size_t len2 = sizeof(data);
       RSA *rsa = EVP_PKEY_get1_RSA(pubkey);
 
-      if(RAND_bytes(secret, sizeof(secret)) == -1) {
+      if(RAND_bytes(secret, sizeof(secret)) <= 0) {
         fprintf(stderr, "error: no randomness.\n");
         ret = false;
         goto decipher_out;
@@ -1660,7 +1776,7 @@ static bool list_readers(ykpiv_state *state) {
 static bool attest(ykpiv_state *state, enum enum_slot slot,
     enum enum_key_format key_format, const char *output_file_name) {
   unsigned char data[2048];
-  unsigned long len = sizeof(data);
+  size_t len = sizeof(data);
   bool ret = false;
   X509 *x509 = NULL;
   int key;
@@ -1682,12 +1798,7 @@ static bool attest(ykpiv_state *state, enum enum_slot slot,
 
   if(key_format == key_format_arg_PEM) {
     const unsigned char *ptr = data;
-    int len2 = len;
-    x509 = X509_new();
-    if(!x509) {
-      fprintf(stderr, "Failed allocating x509 structure.\n");
-      goto attest_out;
-    }
+    int len2 = (int)len;
     x509 = d2i_X509(NULL, &ptr, len2);
     if(!x509) {
       fprintf(stderr, "Failed parsing x509 information.\n");
@@ -1713,7 +1824,7 @@ static bool write_object(ykpiv_state *state, int id,
     const char *input_file_name, int verbosity, enum enum_format format) {
   bool ret = false;
   FILE *input_file = NULL;
-  unsigned char data[3072];
+  unsigned char data[YKPIV_OBJ_MAX_SIZE];
   size_t len = sizeof(data);
   ykpiv_rc res;
 
@@ -1733,7 +1844,7 @@ static bool write_object(ykpiv_state *state, int id,
   }
 
   if(verbosity) {
-    fprintf(stderr, "Writing %lu bytes of data to object %x.\n", len, id);
+    fprintf(stderr, "Writing %lu bytes of data to object %x.\n", (long unsigned int)len, id);
   }
 
   if((res = ykpiv_save_object(state, id, data, len)) != YKPIV_OK) {
@@ -1752,7 +1863,7 @@ write_out:
 static bool read_object(ykpiv_state *state, int id, const char *output_file_name,
     enum enum_format format) {
   FILE *output_file = NULL;
-  unsigned char data[3072];
+  unsigned char data[YKPIV_OBJ_MAX_SIZE];
   unsigned long len = sizeof(data);
   bool ret = false;
 
@@ -1804,6 +1915,7 @@ int main(int argc, char *argv[]) {
               cmdline_parser_action_values[action]);
           return EXIT_FAILURE;
         }
+        /* fall through */
       case action_arg_generate:
       case action_arg_importMINUS_key:
       case action_arg_importMINUS_certificate:
@@ -1856,7 +1968,7 @@ int main(int argc, char *argv[]) {
   }
 
   if(ykpiv_connect(state, args_info.reader_arg) != YKPIV_OK) {
-    fprintf(stderr, "Failed to connect to reader.\n");
+    fprintf(stderr, "Failed to connect to yubikey.\nTry removing and reconnecting the device.");
     return EXIT_FAILURE;
   }
 
@@ -1875,6 +1987,7 @@ int main(int argc, char *argv[]) {
           }
           password = pwbuf;
         }
+        /* fall through */
       case action_arg_generate:
       case action_arg_setMINUS_mgmMINUS_key:
       case action_arg_pinMINUS_retries:
@@ -1941,7 +2054,9 @@ int main(int argc, char *argv[]) {
 
 
   /* openssl setup.. */
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
   OpenSSL_add_all_algorithms();
+#endif
 
 
   for(i = 0; i < args_info.action_given; i++) {
@@ -1990,7 +2105,7 @@ int main(int argc, char *argv[]) {
             fprintf(stderr, "Successfully set new management key.\n");
           }
         } else {
-          fprintf(stderr, "The new management key has to be exactly %d character.\n", KEY_LEN * 2);
+          fprintf(stderr, "The new management key has to be exactly %d hexidecimal characters.\n", KEY_LEN * 2);
           ret = EXIT_FAILURE;
         }
         break;
@@ -2182,6 +2297,8 @@ int main(int argc, char *argv[]) {
   }
 
   ykpiv_done(state);
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
   EVP_cleanup();
+#endif
   return ret;
 }
