@@ -42,7 +42,7 @@
    PURPOSE.  */
 
 /* Please submit changes back to the Scute project at
-   http://www.scute.org/ (or send them to marcus@g10code.com), so that
+   https://www.scute.org/ (or send them to marcus@g10code.com), so that
    they can be picked up by other projects from there as well.  */
 
 /* This file is a modified implementation of the PKCS #11 standard by
@@ -208,10 +208,10 @@ extern "C" {
 #define ck_unlockmutex_t CK_UNLOCKMUTEX
 
 #define ck_c_initialize_args _CK_C_INITIALIZE_ARGS
-#define create_mutex CreateMutex
-#define destroy_mutex DestroyMutex
-#define lock_mutex LockMutex
-#define unlock_mutex UnlockMutex
+#define create_mutex pfnCreateMutex
+#define destroy_mutex pfnDestroyMutex
+#define lock_mutex pfnLockMutex
+#define unlock_mutex pfnUnlockMutex
 #define reserved pReserved
 
 #endif	/* CRYPTOKI_COMPAT */
